"""
PDF Report Generator
--------------------
PDF report creator. Its a postprocess that generates a PDF report with
the model forecasts.
"""
from datetime import datetime
import logging
from pathlib import Path
from typing import Dict

import jupytext
from nbconvert import PDFExporter
import pandas as pd
import papermill as pm
from pkg_resources import resource_filename

from soam.core import Step

logger = logging.getLogger(__name__)


class PDFReport:
    """
    Generate PDF Report object from IPython Notebook.
    """

    def __init__(self, base_path: str):
        """
        Initialization of the PDF Report object.

        Parameters
        ----------
        base_path: str
            string of the base path of the pdf report.
        """
        self.base_path = Path(base_path)

    def export_notebook_to_pdf(self, nb_path: str, nb_params: Dict) -> str:  # type: ignore
        """
        Run notebook, convert, save and return PDF path.

        Parameters
        ----------
        nb_path : str
            Path of the Notebook or Script to Execute .ipynb / .py
        parameters : Dict
            Parameters to run the notebook with (Papermill).

        Returns
        -------
        path
            Generated PDF Path
        """
        report_file = Path(nb_path)

        if not report_file.is_file():
            raise ValueError("Notebook path does not point to a file.")

        nb_params = self._parse_params(nb_params)

        report_filename, report_extension = report_file.stem, report_file.suffix

        if report_extension != "ipynb":
            report_nb = self.base_path / f"{report_filename}.ipynb"
        else:
            report_nb = report_file

        now_timestamp = datetime.today().strftime('%Y-%m-%dT%H-%M-%S')

        run_nb = self.base_path / f"{report_filename} {now_timestamp}.ipynb"
        pdf_filename = self.base_path / f"{report_filename} {now_timestamp}.pdf"

        logger.info("Running %s and converting to %s", run_nb, pdf_filename)

        jupytext_notebook = jupytext.read(report_file)
        # write to execution nb
        jupytext.write(jupytext_notebook, report_nb)

        _ = pm.execute_notebook(str(report_nb), str(run_nb), parameters=nb_params,)

        pdfexp = PDFExporter(
            template_file=resource_filename("soam", "resources/pdf_report.tpl")
        )

        pdf_data, _ = pdfexp.from_filename(run_nb)

        with open(pdf_filename, "wb") as f:
            f.write(pdf_data)

        logger.info("Succesfully wrote: %s", str(pdf_filename))

        return str(pdf_filename)

    def _parse_params(self, nb_params):
        """
        Convert common object types to appropiate parameter types.
        """
        for key, value in nb_params.items():
            if isinstance(value, pd.DataFrame):
                nb_params[key] = value.to_csv(index=False)
        return nb_params


class PDFReportTask(Step, PDFReport):
    """
    Creates the task to generate a PDF report.
    """

    def __init__(self, base_path: str):
        Step.__init__(self)
        PDFReport.__init__(self, base_path)

    def run(self, nb_path: str, nb_params: Dict) -> str:  # type: ignore[override]
        """
        Task to export the notebook into a pdf.

        Parameters
        ----------
        nb_path : str
            Path of the Notebook or Script to Execute .ipynb / .py
        parameters : Dict
            Parameters to run the notebook with (Papermill).

        Returns
        -------
        path
            Generated PDF Path
        """
        return self.export_notebook_to_pdf(nb_path, nb_params)



"""Database saver."""
import logging

from muttlib.dbconn import BaseClient
from muttlib.utils import hash_str
from prefect import Task, context
from prefect.engine.state import State

from soam.core import SoamFlow
from soam.data_models import Base, ForecastValues, SoamFlowRunSchema, SoamTaskRunSchema
from soam.savers import Saver
from soam.utilities.helpers import session_scope

logger = logging.getLogger(__name__)


class DBSaver(Saver):
    def __init__(self, base_client: BaseClient):
        """
        Create a DBSaver object and check if the database
        contains the tables it expect to store the data

        You can use a DBSaver for storing runs or runs + values.
        The values use the runs data as a Foreign Key.

        Parameters
        ----------
        base_client
            A BaseClient with connection to the database
        """
        super().__init__()
        self.db_client = base_client

        if base_client._connect() is not None:
            if not base_client.get_engine().has_table(ForecastValues.__tablename__):
                logger.error(
                    "There are no tables on this database"
                    "Please run: alembic revision --autogenerate "
                    "alembic upgrade head"
                )

    def save_forecast(self, task: Task, old_state: State, new_state: State) -> State:
        """
        Store the forecaster data in the constructed database.

        Parameters
        ----------
        task: Task
            Specify the forecast task you want to save information of.
        old_state: State
            Task old state.
        new_state : State
            Task new state.

        Returns
        -------
        State
            The new updated state of the forecast task.
        """
        if new_state.is_successful():
            save_prediction = new_state.result[0].copy()
            save_prediction["task_run_id"] = context["task_run_id"]
            self.db_client.insert_from_frame(
                save_prediction, ForecastValues.__tablename__
            )

        return new_state

    def save_task_run(self, task: Task, old_state: State, new_state: State) -> State:
        """
        Store the task run information in the database.

        Parameters
        ----------
        task: Task
            Specify the task you want to save information of.
        old_state: State
            Task old state.
        new_state : State
            Task new state.

        Returns
        -------
        State
            The new updated state of the task.
        """
        if new_state.is_running():
            flow_run_id = context["flow_run_id"]
            task_run_id = context["task_run_id"]

            step_type = self.get_task_type(task)

            insert_fr = SoamTaskRunSchema(
                task_run_id=task_run_id,
                params=repr(task),
                params_hash=hash_str(repr(task)),
                step_type=step_type,
                flow_run_id=flow_run_id,
            )
            _ = self._insert_single(insert_fr)

        return new_state

    def save_flow_run(
        self, soamflow: SoamFlow, old_state: State, new_state: State
    ) -> State:
        """
        Save the SoamFlow run data in the create connection to a database.

        Parameters
        ----------
        saomflow: SoamFlow
            Specify the soamflow you want to save information of.
        old_state: State
            SoamFlow old state.
        new_state : State
            SoamFlow new state.

        Returns
        -------
        State
            The new updated state of the SoamFlow.
        """
        if new_state.is_running():
            flow_run_id = context["flow_run_id"]
            run_date = context["date"]

            insert_fr = SoamFlowRunSchema(
                flow_run_id=flow_run_id,
                run_date=run_date,
                start_datetime=soamflow.start_datetime,
                end_datetime=soamflow.end_datetime,
            )

            _ = self._insert_single(insert_fr)

        return new_state

    def _insert_single(self, element: Base) -> int:
        with session_scope(engine=self.db_client.get_engine()) as session:
            session.add(element)  # pylint: disable=maybe-no-member

        return element



"""
Store results
----------
A class to store results in a table
"""

from typing import Dict

from muttlib.dbconn import BaseClient
import pandas as pd

from soam.core import Step


class Store(Step):
    def __init__(
        self, db_cli: BaseClient, table: str, extra_insert_args: Dict = None, **kwargs
    ):
        """Store given data into a DataBase

        Parameters
        ----------
        db_cli:
            BaseClient client.
        table:
            str of table to store in.
        extra_insert_args:
            dict extra arguments to insert data.
        """
        super().__init__(**kwargs)

        self.db_cli = db_cli
        self.table = table
        self.extra_insert_args = extra_insert_args

    def run(self, df: pd.DataFrame):  # type: ignore
        """
        Store given data frame

        Parameters
        ----------
        df
            A pandas DataFrame to store.
        """
        if not self.extra_insert_args:
            self.extra_insert_args = {}

        return self.db_cli.insert_from_frame(
            df=df, table=self.table, **self.extra_insert_args
        )



"""SoaM reporting."""

from soam.reporting.gsheets_report import GSheetsReportTask
from soam.reporting.mail_report import MailReportTask
from soam.reporting.pdf_report import PDFReportTask
from soam.reporting.slack_report import SlackReportTask



# step.py
"""
Step Base Class
---------------
An abstract base class for the steps in the pipeline, including: postprocess,
preprocess, extract and forecaster.
"""
from abc import abstractmethod

import pandas as pd
from prefect import Task, context
from sklearn.base import BaseEstimator


class Step(Task, BaseEstimator):
    """
    The base class for all steps.
    All implementations of step have to implement the `run()` method defined below.
    """

    @abstractmethod
    def run(self) -> pd.DataFrame:
        """
        Execute this step function

        Parameters
        ----------
        time_series
            A pandas DataFrame containing the data for the step

        Returns
        -------
        pandas.DataFrame
            pandas DataFrame containing the output values.
        """

    def get_task_id(self) -> str:  # pylint: disable=no-self-use
        """
        Obtains the task id.

        Returns
        -------
        flow_run_id
            the id of the flow run.
        """
        return context["flow_run_id"]

    def __repr__(self) -> str:
        return f"<Task: {BaseEstimator.__repr__(self)}>"



"""Orbit model wrapper."""
from typing import List, Union
import warnings

from orbit.models.dlt import DLTFull
import pandas as pd
from typing_extensions import Literal

from soam.constants import SEED
from soam.models._base import SkWrapper, sk_constructor_wrapper
from soam.utilities.utils import SuppressStdOutStdErr

# pylint: disable=super-init-not-called, attribute-defined-outside-init, unnecessary-pass, no-member


class SkOrbit(SkWrapper):
    """Scikit-Learn Orbit model wrapper."""

    _ignore_params = ["full_output"]

    @sk_constructor_wrapper(DLTFull)
    def __init__(
        self,
        date_col: str = 'date',
        response_col: str = None,
        regressor_col: Union[List[str], str] = None,
        damped_factor: float = 0.8,
        period: int = 1,
        seasonality: int = -1,
        seed: int = SEED,
        chains: int = 1,
        global_trend_option: Literal[
            'flat', 'linear', 'loglinear', 'logistic'
        ] = "linear",
        full_output: bool = False,
    ):
        """Constructor with extra parameters in addition to the base model ones.

        Parameters
        ----------
        response_col : str
            response or y column name
        date_col : str
            date column name
        regressor_col : Union[List[str], str]
            extra regressors column names
        damped_factor : float, optional
            by default 0.8
        period : int, optional
            by default 1
        seasonality : int, optional
            by default -1
        chains : int, optional
            number of chains spawned by PyStan, by default 1
        seed : int, optional
            by default 1
        global_trend_option : Literal[, optional
            by default "linear"
        full_output : bool, default False
            Return full Orbit output or just prediction column.

        Notes:
            Since Orbit manages kwargs for everything, this is awkward for our
            constructor patching strategy, i.e we need explicit arguments in the signature.
            That's why we explicitly specify them in the wrapper's signature,
            and we mark custom params in _to_ignore.

            For more details on model specific parameters check docs.
        """
        pass

    def fit(self, X: pd.DataFrame, y: pd.Series):
        """Fit estimator to data."""
        df = self._transform_to_input_format(X, y)
        self.model = self._init_sk_model(DLTFull, ignore_params=self._ignore_params)
        with warnings.catch_warnings(), SuppressStdOutStdErr():
            warnings.simplefilter("ignore")
            self.model.fit(df)
        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """Scikit learn's predict."""
        X = self._transform_to_input_format(X)

        predictions = self.model.predict(X)  # pylint: disable=assignment-from-no-return
        predictions = self._transform_to_output_format(predictions)

        return predictions

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Scikit learn's transform."""
        return self.predict(X)

    def fit_transform(self, X: pd.DataFrame, y: pd.Series):
        """Scikit learn's fit_transform."""
        self.fit(X, y)
        return self.transform(X)

    def _transform_to_input_format(
        self, X: pd.DataFrame, y: pd.Series = None
    ) -> pd.DataFrame:
        """Transform input to Orbit compatible df."""
        if y is not None:
            # set response col dynamically
            self.response_col = y.name
            return X.assign(**{self.response_col: y})
        return X

    def _transform_to_output_format(self, predictions: pd.Series) -> pd.DataFrame:
        """Transform Orbit output to SoaM format."""
        predictions = predictions.rename(columns={"prediction": "yhat"})
        if self.full_output:
            return predictions
        return predictions[[self.date_col, "yhat"]]



"""statsmodels.holtwinters estimators."""
import logging
from typing import Dict, Tuple

import pandas as pd
from statsmodels.tsa.holtwinters import ExponentialSmoothing

from soam.constants import DS_COL, YHAT_COL
from soam.models._base import SkWrapper, sk_constructor_wrapper
from soam.utilities.utils import SuppressStdOutStdErr

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# pylint: disable=super-init-not-called, attribute-defined-outside-init, unnecessary-pass, no-member


class SkExponentialSmoothing(SkWrapper):
    """Scikit-Learn statsmodels.ExponentialSmoothing model wrapper."""

    @sk_constructor_wrapper(ExponentialSmoothing)
    def __init__(
        self, fit_params: Dict = None, date_col: str = DS_COL,
    ):
        """Construct wrapper with extra parameters in addition to the ExponentialSmoothing ones.

        Parameters
        ----------
        fit_params : Dict, optional
            Params to be fed to model's .fit() method, by default None
        date_col : str, optional
            date column, by default None
        """
        pass

    def fit(self, X: pd.DataFrame, y: pd.Series):
        """
        Fit estimator to data.

        Notes:
            Since ExponentialSmoothing requires endog arrays (not colnames) at
            init time, we transform the input before _init_sk_model.
        """
        # arrays are required at model initialization
        self.endog = self._transform_to_input_format(X, y)
        self.model = self._init_sk_model(ExponentialSmoothing, clean=True)
        with SuppressStdOutStdErr():

            self.model_fit = (
                self.model.fit()
                if self.fit_params is None
                else self.model.fit(**self.fit_params)
            )
        # saving as prediction start point
        self._train_len = len(X)

        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """Scikit learn's predict."""
        X_len, _ = self._transform_to_input_format(X)
        start = self._train_len
        end = start + X_len - 1
        predictions = self.model_fit.predict(start, end)
        predictions = self._transform_to_output_format(predictions, X)
        return predictions

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Scikit learn's transform."""
        return self.predict(X)

    def fit_transform(self, X: pd.DataFrame, y: pd.Series, output_length: int = 1):
        """Scikit learn's fit_transform with output_length."""
        X_train = X[:-output_length]
        X_pred = X[-output_length:]
        y_train = y[:-output_length]
        self.fit(X_train, y_train)
        return self.transform(X_pred)

    def _transform_to_input_format(
        self, X: pd.DataFrame, y: pd.Series = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Transform input to ExponentialSmoothing compatible format."""
        # saving as prediction start point
        if y is not None:
            # train
            if self.date_col is not None:
                return pd.Series(y.values, index=X[self.date_col].values)
            return pd.Series(y.values, index=X.index)
        else:
            # predict
            return X.pipe(len), y

    def _transform_to_output_format(
        self, predictions: pd.Series, X_pred: pd.DataFrame
    ) -> pd.DataFrame:
        """Transform ExponentialSmoothing output to SoaM format."""
        final_predictions = pd.DataFrame()
        final_predictions[self.date_col] = X_pred[self.date_col].values
        final_predictions[YHAT_COL] = predictions.values
        return final_predictions



"""Workflow backtester."""
from collections.abc import Mapping
import logging
from typing import (  # pylint:disable=unused-import
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import pandas as pd
from prefect.utilities.tasks import defaults_from_attrs

from soam.constants import DS_COL, Y_COL, YHAT_COL
from soam.core import Step
from soam.utilities.utils import add_future_dates, split_backtesting_ranges
from soam.workflow.forecaster import Forecaster
from soam.workflow.transformer import DummyDataFrameTransformer, Transformer

if TYPE_CHECKING:
    from soam.plotting import ForecastPlotterTask
    from soam.savers import Saver


logger = logging.getLogger(__name__)
RANGES_KEYWORD = "ranges"
METRICS_KEYWORD = "metrics"
PLOT_KEYWORD = "plot"
DEFAULT_METRIC_AGGREGATION = {
    "avg": lambda metric_values: sum(metric_values) / len(metric_values),
    "max": max,
    "min": min,
}


class Backtester(Step):
    """
    Class to perform backtesting.

    Note: To run a single fold backtest, for example to validate the model
    performance in the last run, pass a timeseries with the exact lenght of
    train_window plus test_window.

    Parameters
    ----------
    forecaster : soam.Forecaster
        Forecaster that will be fitted and execute the predictions.
    test_window: pd.Timedelta
        Time range to be extracted from the main timeseries on which the model will be evaluated on each backtesting run.
        If `None` then `forecaster.output_length` is be used.
    train_window: pd.Timedelta, optional
        Time range on which the model will trained on each backtesting run.
        If a pd.Timedelta value is passed then the sliding method will be used to select the training data.
        If `None` then the full time series will be used. This is the expanding window method.
    step_size: int
        Distance between each successive step between the beginning of each forecasting
        range. If None defaults to test_window.
    metrics: dict(str, callable)
        `dict` containing name of a metric and a callable to compute it.
        The callable must conform to the interface used by sklearn for regression metrics:
        https://scikit-learn.org/stable/modules/classes.html#regression-metrics
    savers : list of soam.savers.Saver, optional
        The saver to store the parameters and state changes.
    aggregation: bool or dict
        The expected aggregations for the results.
        If set to true will use the default aggregation, this keeps the last plot,
        and calculates the average, minimum and maximum for the different metrics.
        If it's a dict the PLOT_KEYWORD is expected to be assigned to the index
        of the selected plot. METRICS_KEYWORD is expected to be another dictionary
        containing the name of the aggregation associated with the function to
        aggregate de list of values per metric.
        If aggregation is set to False or None, no aggregation would be performed.
        #TODO: make PLOT_KEYWORD support tuples to pick slices.
    """

    def __init__(
        self,
        forecaster: "Forecaster",
        preprocessor: "Transformer" = None,
        forecast_plotter: "ForecastPlotterTask" = None,
        test_window: "Optional[int]" = 1,
        train_window: "Optional[int]" = 1,
        step_size: "Optional[int]" = None,
        metrics: "Dict[str, Callable]" = None,
        savers: "Optional[List[Saver]]" = None,
        aggregation: Union[str, Dict] = None,
        **kwargs,
    ):
        """
        Backtester object initialization.

        Parameters
        ----------
            forecaster: soam.Forecaster
                Forecaster that will be fitted and execute the predictions.
            preprocessor: soam.Transformer
                Transformer that will be used to preprocess the data, optional and defaults to None.
            forecast_plotter: soam.Plotting.ForecastPlotterTask
                Forecast plotter, optional and defaults to None.
            test_window: pd.Timedelta
                Time range to be extracted from the main timeseries on which the model will be evaluated on each backtesting run.
                If `None` then `forecaster.output_length` is be used.
            train_window: pd.Timedelta
                Time range on which the model will trained on each backtesting run.
                If a pd.Timedelta value is passed then the sliding method will be used to select the training data.
                If `None` then the full time series will be used. This is the expanding window method.
            step_size: int
                Distance between each successive step between the beginning of each forecasting
                range. If None defaults to test_window.
            metrics: dict(str, callable)
                `dict` containing name of a metric and a callable to compute it.
                The callable must conform to the interface used by sklearn for regression metrics:
                https://scikit-learn.org/stable/modules/classes.html#regression-metrics
            savers: list of soam.savers.Saver, optional
                The saver to store the parameters and state changes.
            aggregation: bool or dict
                The expected aggregations for the results.
                If set to true will use the default aggregation, this keeps the last plot,
                and calculates the average, minimum and maximum for the different metrics.
                If it's a dict the PLOT_KEYWORD is expected to be assigned to the index
                of the selected plot. METRICS_KEYWORD is expected to be another dictionary
                containing the name of the aggregation associated with the function to
                aggregate de list of values per metric.
                If aggregation is set to False or None, no aggregation would be performed.
        """
        super().__init__(**kwargs)
        if savers is not None:
            for saver in savers:  # pylint: disable=unused-variable
                pass
                # self.state_handlers.append(saver.save_step)

        if preprocessor is None:
            preprocessor = DummyDataFrameTransformer()

        self.forecaster = forecaster
        self.preprocessor = preprocessor
        self.forecast_plotter = forecast_plotter
        self.test_window = test_window
        self.train_window = train_window
        self.step_size = step_size
        self.metrics = metrics
        self.aggregation = aggregation

    @defaults_from_attrs(
        'forecaster',
        'preprocessor',
        'forecast_plotter',
        'test_window',
        'train_window',
        'step_size',
        'metrics',
        'aggregation',
    )
    def run(  # type: ignore
        self,
        time_series: pd.DataFrame,
        forecaster: Forecaster = None,
        preprocessor: Transformer = None,
        forecast_plotter: "ForecastPlotterTask" = None,
        test_window: pd.Timedelta = None,
        train_window: Optional[int] = None,
        step_size: Optional[int] = None,
        metrics: Dict[str, Callable] = None,
        aggregation: Union[bool, Dict] = None,
    ) -> List[Dict[str, Any]]:
        """
        Train the model with past data and compute metrics.

        Parameters
        ----------
        time_series: pd.DataFrame
            Data used to train and evaluate the data.
        forecaster : soam.Forecaster
            Forecaster that will be fitted and execute the predictions.
        preprocessor: Transformer
            Provide an interface to transform pandas DataFrames.
        forecast_plotter: ForecastPlotterTask
            Plot forecasts.
        test_window: pd.Timedelta
            Time range to be extracted from the main timeseries on which the model will
            be evaluated on each backtesting run.
            If `None` then `forecaster.output_length` is be used.
        train_window: pd.Timedelta, optional
            Time range on which the model will trained on each backtesting run.
            If a pd.Timedelta value is passed then the sliding method will be used to
            select the training data.
            If `None` then the full time series will be used. This is the expanding
            window method.
        step_size: int
            Distance between each successive step between the beginning of each
            forecasting
            range. If None defaults to test_window.
        metrics: dict(str, callable)
            `dict` containing name of a metric and a callable to compute it.
            The callable must conform to the interface used by sklearn for regression
            metrics:
            https://scikit-learn.org/stable/modules/classes.html#regression-metrics
        aggregation: bool or dict
            The expected aggregations for the results.
            If set to true will use the default aggregation, this keeps the last plot,
            and calculates the average, minimum and maximum for the different metrics.
            If it's a dict the PLOT_KEYWORD is expected to be assigned to the index
            of the selected plot. METRICS_KEYWORD is expected to be another dictionary
            containing the name of the aggregation associated with the function to
            aggregate de list of values per metric.
            If aggregation is set to False or None, no aggregation would be performed.
            #TODO: make PLOT_KEYWORD support tuples to pick slices.
        """
        # TODO
        # - What is the effect of reusing steps like this if they have saver set?

        if test_window is None:
            test_window = forecaster.output_length  # type: ignore

        time_series_splits = split_backtesting_ranges(
            time_series, test_window, train_window, step_size,
        )
        rv = []
        for train_set, test_set in time_series_splits:
            slice_rv = {}

            fc = forecaster.copy()  # type: ignore
            preproc = preprocessor.copy()  # type: ignore

            ready_train_set, fitted_preproc = preproc.run(train_set)
            ready_train_set = add_future_dates(ready_train_set, periods=test_window)
            prediction, _, _ = fc.run(ready_train_set)
            train_start = train_set[DS_COL].min()
            train_end = train_set[DS_COL].max()
            test_end = test_set[DS_COL].max()
            slice_rv[RANGES_KEYWORD] = (train_start, train_end, test_end)

            ready_test_set = fitted_preproc.transform(test_set)
            slice_metrics = compute_metrics(
                ready_test_set[Y_COL], prediction[YHAT_COL], metrics
            )
            slice_rv[METRICS_KEYWORD] = slice_metrics

            if forecast_plotter:
                full_set = pd.concat([ready_train_set, ready_test_set])
                fcp = forecast_plotter.copy()
                fcp.path = (
                    fcp.path.parent
                    / f"train_start={train_start}_train_end={train_end}_test_end={test_end}_{fcp.path.name}"
                )
                slice_rv[PLOT_KEYWORD] = fcp.run(full_set, prediction)

            rv.append(slice_rv)

        if aggregation:
            return aggregate_rv(aggregation, rv)
        return rv


def aggregate_rv(
    aggregation: Union[bool, Dict], result_values: List[Dict[str, Any]],
) -> List[Dict[str, Any]]:
    """
    Aggregates a list of results from the Backtester.

    Parameters
    ----------
    aggregation: bool or dict
        The expected aggregations for the results.
        If set to true will use the default aggregation, this keeps the last plot,
        and calculates the average, minimum and maximum for the different metrics.
        If it's a dict the PLOT_KEYWORD is expected to be assigned to the index
        of the selected plot. METRICS_KEYWORD is expected to be another dictionary
        containing the name of the aggregation associated with the function to
        aggregate de list of values per metric.
        #TODO: make PLOT_KEYWORD support tuples to pick slices.
    result_values: list of dict of str and any
        List containing the results of the different slices of the backtester.

    Returns
    -------
        #TODO: check if we need to return list for interface compatibility or we can
        #TODO: return just the dict.
        list of dict of str and any
        A list containing one dict with the whole range of the splits, the selected
        plot and the different aggregation functions per metric.
    """
    metric_aggregation = DEFAULT_METRIC_AGGREGATION
    aggregated_plot = result_values[-1][PLOT_KEYWORD]
    if isinstance(aggregation, Mapping):
        if METRICS_KEYWORD in aggregation:
            metric_aggregation = aggregation[METRICS_KEYWORD]
        if PLOT_KEYWORD in aggregation:
            aggregated_plot = result_values[aggregation[PLOT_KEYWORD]][PLOT_KEYWORD]

    metrics_to_aggregate: Dict[str, List] = {}
    for split_result in result_values:
        for metric, value in split_result[METRICS_KEYWORD].items():
            metrics_to_aggregate.setdefault(metric, []).append(value)

    aggregated_metrics = {
        metric: {
            aggregation_name: metric_aggregation[aggregation_name](  # type: ignore
                metrics_to_aggregate[metric]
            )
            for aggregation_name in metric_aggregation.keys()
        }
        for metric in metrics_to_aggregate.keys()
    }

    return [
        {
            RANGES_KEYWORD: (
                result_values[0][RANGES_KEYWORD][0],
                result_values[-1][RANGES_KEYWORD][-1],
            ),
            METRICS_KEYWORD: aggregated_metrics,
            PLOT_KEYWORD: aggregated_plot,
        }
    ]


def compute_metrics(y_true, y_pred, metrics):
    """
    Computation of a given metric

    Parameters
    ----------
    y_true:
        True values for y
    y_pred:
        Predicted values for y.
    metrics: metric
        Chosen performance metric to measure the model capabilities.

    Returns
    -------
    dict
        Performance metric and its value"""
    return {metric_name: func(y_true, y_pred) for metric_name, func in metrics.items()}



# helpers.py
# pylint: skip-file
"""
Utility functions for the project.

TODO: review this file, it seems unused in the rest of the project.
"""
from contextlib import contextmanager
from datetime import timedelta
import logging
from pathlib import Path

from muttlib.utils import make_dirs, str_to_datetime
import numpy as np
import pandas as pd
from pandas.tseries import offsets
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

from soam.constants import DS_COL, HOURLY_TIME_GRANULARITY, YHAT_COL
from soam.utilities.utils import range_datetime

logger = logging.getLogger(__name__)

RUN_ID_COL = "run_id"


def get_store_dir(base_dir, kpi, prefix, date, end_date=None, sample_size=None):
    """
    Get parquet cache storage directory.

    Parameters
    ----------
    base_dir: Union[str, Path]
        base directory path
    kpi: str
        key performance indicator being used
    prefix: str
        prefix string
    date: datetime
        datetime
    end_date: datetime
        end datetime
    sample_size: int
        size of the sample

    Returns
    -------
    path
        The new store_dir directory created.
    """
    store_dir = base_dir / kpi / prefix
    store_dir /= f'{date:%Y%m%d}{f"-{end_date:%Y%m%d}" if end_date is not None else ""}'
    if sample_size is not None:
        store_dir /= f"sample-{sample_size}"
    return make_dirs(store_dir)


def get_store_file(
    base_dir,
    kpi,
    date,
    sample_size=None,
    end_date=None,
    prefix=None,
    suffix=None,
    ft="pkl",
    extra_dir=None,
    end_date_hour=False,
):
    """
    Create name to store tabular data query outputs.

    Parameters
    ----------

    base_dir: Union[str, Path]
        base directory path.
    kpi: str
        key performance indicator being used.
    date: datetime
        datetime.
    sample_size: int
        size of the sample
    end_date: datetime
        end datetime.
    prefix: str
        prefix string.
    suffix: str
        suffix string.
    ft: str
        file extension, by default is "pkl" for pickle files.
    extra_dir: str
        extra directory.
    end_date_hour: boolean
        boolean value to determine if the end date hour should be included. False by default.

    Returns
    -------
    path
        path of the store file.
    """
    save_dir = base_dir / kpi
    if extra_dir is not None:
        save_dir /= extra_dir
    save_dir = make_dirs(save_dir)
    outf = f"{date:%Y%m%d}"
    if end_date is not None:
        outf += f"_{end_date:%Y%m%d}"
        if end_date_hour:
            outf += f"T{end_date:%H}"
    if sample_size is not None:
        outf += f"_sample-{sample_size}"
    if prefix is not None:
        outf = f"{prefix}_{outf}"
    if suffix is not None:
        outf = f"{outf}_{suffix}"
    if ft is not None:
        outf += f".{ft}"
    return save_dir / outf


def get_figure_full_path(
    fig_dir,
    target_col,
    fig_name,
    start_date,
    end_date,
    time_granularity,
    granularity,
    suffix=None,
    as_posix=True,
    end_date_hour=False,
):
    """
    Create figure file-path for given model inputs.

    Parameters
    ----------
    fig_dir: path
        figure directory.
    target_col
        target column.
    fig_name: str
        name of the figure
    start_date: date
        start date of analysis.
    end_date: date
        end date of analysis
    time_granularity: str
        time granularity, how much a period represents.
    granularity: str
        granularity.
    suffix: str
        suffix string.
    as_posix: boolean
        True by default.
    end_date_hour: boolean
        boolean value to determine if the end date hour should be included. False by default.

    Returns
    -------
    path
        Path of the created figure.
    """
    save_dir = make_dirs(fig_dir / granularity / time_granularity / fig_name)
    root_name = f"{start_date:%Y%m%d}_{end_date:%Y%m%d}"
    if end_date_hour:
        root_name += f"T{end_date:%H}"
    root_name += f"_{target_col}"
    if suffix is not None:
        root_name += f"_{suffix}"
    fig_path = Path(save_dir, root_name + ".png")
    if as_posix:
        return fig_path.as_posix()
    return fig_path


def create_forecaster_dates(
    end_date, forecaster_train_window, forecaster_future_window
):
    """
    Process and correct all respective dates for forecaster.

    Parameters
    ----------
    end_date: date
        end date of analysis
    forecaster_train_window
        window of the train set to train the forecaster.
    forecaster_future_window
        future time window to predict the target variable value for.

    Returns
    -------
    date
        forecaster dates.
    """
    # Check not both future and end date as nulls, assert order
    train_w = forecaster_train_window
    future_w = forecaster_future_window
    m = f'Future ("{future_w}") or train ("{train_w}") windows are not geq 0.'
    assert all([future_w > 0, train_w >= 0]), m
    end_date = str_to_datetime(end_date) if isinstance(end_date, str) else end_date
    start_date = end_date - offsets.Day(train_w)
    future_date = end_date + offsets.Day(future_w)
    return start_date, end_date, future_date


def create_anomaly_window_dates(end_date, anomaly_window, time_granularity):
    """
    Creation of the anomaly window dates.

    Parameters
    ----------
    end_date: date
        end date of analysis
    anomaly_window
        window time to analyze anomalies.
    time_granularity
        time granularity, how much a period represents.

    Returns
    -------
    list
        anomaly window dates.
    """
    hourly_offset = True if time_granularity == HOURLY_TIME_GRANULARITY else False
    end_date = str_to_datetime(end_date) if isinstance(end_date, str) else end_date
    anomaly_window_start_date = end_date - timedelta(days=(anomaly_window - 1))
    return list(
        range_datetime(
            anomaly_window_start_date,
            end_date,
            hourly_offset=hourly_offset,
            as_datetime=False,
        )
    )


def insert_df_multiple_clients(
    db_clients, insertion_ids, table_name, df, coseries_ids=None, **kwargs
):
    """
    Insert a dataframe to a list of database connections.

    Parameters
    ----------
    db_clients
        clients database
    insertion_ids
        ids to insert data in.
    table_name: str
        name of the table
    df: pd.DataFrame
        pandas dataframe to insert in database.
    coseries_ids=None
    """

    for db_cli in db_clients:
        dialect = db_cli.dialect
        if insertion_ids:
            df["insertion_id"] = insertion_ids[dialect]
        if coseries_ids:
            df["coseries_id"] = coseries_ids[dialect]
        logger.debug(f"Writing to {dialect} {table_name} table...")
        db_cli.insert_from_frame(df, table_name, if_exists="append", **kwargs)
        logger.info(f'Inserted {len(df)} records in {dialect} "{table_name}" table.')
    return


# TODO remove this function when added to muttlib
@contextmanager
def session_scope(engine: Engine, **session_kw):
    """
    Provide a transactional scope around a series of operations.

    Parameters
    ----------
    engine: Engine
        engine connection to the session.
    """

    Session = sessionmaker(bind=engine)
    session = Session(**session_kw)

    try:
        yield session
        session.commit()
    except Exception as err:
        logger.exception(err)
        session.rollback()
        raise
    finally:
        session.close()



"""Integration test module."""
from copy import deepcopy

import pandas as pd
import pytest
from sklearn.metrics import mean_absolute_error, mean_squared_error

from soam.constants import (
    ANOMALY_PLOT,
    DS_COL,
    FIG_SIZE,
    MONTHLY_TIME_GRANULARITY,
    PLOT_CONFIG,
    YHAT_COL,
)
from soam.core import SoamFlow
from soam.models import SkProphet
from soam.plotting import ForecastPlotterTask
from soam.reporting import SlackReportTask
from soam.reporting.slack_report import (
    DEFAULT_FAREWELL_MESSAGE,
    DEFAULT_GREETING_MESSAGE,
)
from soam.savers import CSVSaver
from soam.utilities.utils import add_future_dates
from soam.workflow import Backtester, Forecaster, Transformer
from soam.workflow.backtester import METRICS_KEYWORD, RANGES_KEYWORD
from tests.helpers import sample_data_df  # noqa pylint:disable=unused-import
from tests.workflow.test_backtester import (
    SimpleProcessor,
    assert_backtest_all_folds_result,
)

# pylint:disable=protected-access, redefined-outer-name


def create_slack_message(
    predictions,
    metric,
    farewell_message=DEFAULT_FAREWELL_MESSAGE,
    greeting_message=DEFAULT_GREETING_MESSAGE,
):
    """Function to create the slack message."""
    greeting_message.format(metric_name=metric)

    summary_entries = []
    summary_entries.append(greeting_message)

    for _, row in predictions.iterrows():
        date = row[DS_COL].strftime('%Y-%b-%d')
        value = "{:.2f}".format(row[YHAT_COL])
        summary_entries.append(f"• *[{date}]* {value}\n")

    summary_entries.append(farewell_message)
    return "\n".join(summary_entries)


def create_settings_file(path, content):
    """Function to acquire the settings file path."""
    # Code to acquire resource, e.g.:
    setting_path = path / "settings.ini"
    with open(setting_path, "w+") as f:
        f.write(content)
    return setting_path


def assert_saved_data(path, prefix):
    """Assert savers worked as expected."""
    assert len(list(path.glob(f"{prefix}*"))) == 1


@pytest.fixture
def expected_predictions() -> pd.DataFrame:
    pred = {
        DS_COL: pd.date_range("2016-06-01", periods=7, freq="MS"),
        "yhat": [
            453990.93,
            461908.61,
            482625.56,
            438345.63,
            458133.30,
            467090.52,
            508706.42,
        ],
    }
    return pd.DataFrame(pred)


def test_integration_forecast(
    sample_data_df, expected_predictions, tmp_path  # noqa: F811
):
    """Test simple prediction workflow."""
    # TODO: Test adding saver to flow
    test_run_name = "test_integration_forecasting"
    saver_data = CSVSaver(tmp_path)
    my_model = SkProphet()
    sample_data_df = add_future_dates(sample_data_df, 7)
    forecaster = Forecaster(my_model, output_length=7, savers=[saver_data])
    with SoamFlow(name=test_run_name, saver=None) as f:  # noqa
        preds_model = forecaster(time_series=sample_data_df)

    fs = f.run()

    assert 3 == len(fs.result[preds_model]._result.value)
    assert expected_predictions.ds.equals(fs.result[preds_model]._result.value[0].ds)
    assert expected_predictions.yhat.values == pytest.approx(
        fs.result[preds_model]._result.value[0].yhat, 0.1
    )
    assert fs.is_successful()
    assert_saved_data(tmp_path, test_run_name)


def test_integration_backtest(
    tmp_path, sample_data_df  # noqa: F811
):  # pylint: disable=redefined-outer-name
    """Test backtest run."""
    # TODO: Add savers assertions once integrated with Backtest
    test_run_name = "test_integration_backtesting"
    test_window = 30
    train_data = pd.concat([sample_data_df] * 3)
    train_data[DS_COL] = pd.date_range(
        train_data[DS_COL].min(), periods=len(train_data), freq='MS'
    )
    model = SkProphet()
    forecaster = Forecaster(model=model, output_length=test_window)
    plot_config = deepcopy(PLOT_CONFIG)
    plot_config[ANOMALY_PLOT][MONTHLY_TIME_GRANULARITY][FIG_SIZE] = (8, 3)
    forecast_plotter = ForecastPlotterTask(
        path=tmp_path,
        metric_name='test',
        time_granularity=MONTHLY_TIME_GRANULARITY,
        plot_config=plot_config,
    )
    metrics = {
        "mae": mean_absolute_error,
        "mse": mean_squared_error,
    }

    backtester = Backtester(
        forecaster=forecaster,
        preprocessor=Transformer(SimpleProcessor()),
        forecast_plotter=forecast_plotter,
        test_window=test_window,
        train_window=30,
        metrics=metrics,
        aggregation="default",
    )

    with SoamFlow(name=test_run_name, saver=None) as f:  # noqa
        backtest = backtester(train_data)

    fs = f.run()

    expected_values = [
        {
            RANGES_KEYWORD: (
                pd.Timestamp('2013-02-01 00:00:00'),
                pd.Timestamp('2023-01-01 00:00:00'),
            ),
            METRICS_KEYWORD: {
                'mae': {
                    'avg': 1.9589288335569606,
                    'max': 3.1358162976127217,
                    'min': 1.140921182444867,
                },
                'mse': {
                    'avg': 6.503755107101683,
                    'max': 12.666965373730687,
                    'min': 2.4605768804352675,
                },
            },
            'plots': '0_forecast_2018020100_2020080100_.png',
        }
    ]
    rvs = fs.result[backtest]._result.value
    assert_backtest_all_folds_result(rvs, expected_values)


def test_integration_forecast_and_report(
    mocker, sample_data_df, expected_predictions, tmp_path  # noqa: F811
):  # pylint: disable=redefined-outer-name
    """Test prediction workflow with reporting."""
    # TODO: Test adding saver to flow
    test_run_name = "test_integration_forecast_and_report"
    saver_data = CSVSaver(tmp_path)
    my_model = SkProphet()
    sample_data_df = add_future_dates(sample_data_df, 7)
    forecaster = Forecaster(my_model, output_length=7, savers=[saver_data])

    setting_path = create_settings_file(tmp_path, "[settings]\nSLACK_TOKEN=token")
    slack_report = SlackReportTask("channel_id", "mae", setting_path)
    slack_report.slack_client.files_upload = mocker.stub(name="files_upload")

    plot_config = deepcopy(PLOT_CONFIG)
    plot_config[ANOMALY_PLOT][MONTHLY_TIME_GRANULARITY][FIG_SIZE] = (8, 3)
    forecast_plotter = ForecastPlotterTask(
        path=tmp_path,
        metric_name='test',
        time_granularity=MONTHLY_TIME_GRANULARITY,
        plot_config=plot_config,
    )

    with SoamFlow(name=test_run_name, saver=None) as f:  # noqa
        preds_model = forecaster(time_series=sample_data_df)
        plot_fn = forecast_plotter(preds_model[1], preds_model[0])
        slack_report(preds_model[0], plot_fn)

    fs = f.run()
    assert 3 == len(fs.result[preds_model]._result.value)
    assert expected_predictions.ds.equals(fs.result[preds_model]._result.value[0].ds)
    assert expected_predictions.yhat.values == pytest.approx(
        fs.result[preds_model]._result.value[0].yhat, 0.1
    )
    assert fs.is_successful()
    plot_filename = str(fs.result[plot_fn]._result.value)
    # expected message built with actual preds to avoid rounding issues
    expected_msg = create_slack_message(fs.result[preds_model]._result.value[0], "mae")
    slack_report.slack_client.files_upload.assert_called_once_with(
        channels="channel_id",
        file=plot_filename,
        initial_comment=expected_msg,
        title="mae Forecast",
    )
    assert_saved_data(tmp_path, test_run_name)



# slack_report.py
"""
Slack Report
------------
Slack reporting and message formatting tools. Its a postprocess that sends the
model forecasts though the slack app.
"""
from asyncio import Future
from pathlib import Path
from typing import Any, Optional, Union

import pandas as pd
import slack
from slack.web.slack_response import SlackResponse

from soam.cfg import get_slack_cred
from soam.constants import DS_COL, YHAT_COL
from soam.core.step import Step

DEFAULT_GREETING_MESSAGE = "Hello everyone! Here are the results of the forecast for the *{metric_name}* metric:\n"
DEFAULT_FAREWELL_MESSAGE = "Cheers!\n SoaM."


class SlackReport:
    """
    Generates the report to share via Slack.
    """

    def __init__(
        self, channel_id: str, metric_name: str, setting_path: Optional[str],
    ):
        """
        Initialization of the Slack Report object.

        Parameters
        ----------
        channel_id: str
            Slack channel id where the report will be sent.
        metric_name: str
            Performance metric being measured.
        setting_path: str
            Setting path.
        """
        credentials = get_slack_cred(setting_path)
        self.slack_client = slack.WebClient(credentials["token"])
        self.channel_id = channel_id
        self.metric_name = metric_name

    def send_report(
        self,
        prediction: pd.DataFrame,
        plot_filename: Union[str, Path],
        greeting_message: Optional[str] = DEFAULT_GREETING_MESSAGE,
        farewell_message: Optional[str] = DEFAULT_FAREWELL_MESSAGE,
    ) -> Union[Future, SlackResponse]:
        """
        Send Slack report.

        Parameters
        ----------
        prediction : pd.DataFrame
            DataDrame of the predictions made.
        plot_filename : str or pathlib.Path
             Path of the forecast data to send.
        greeting_message: str
            Greeting message to send via Slack with the predictions.
        farewell_message: str
            Farewell message to send via Slack with the predictions.

        Returns
        -------
        Slack Report
            Sends the specified message with the predictions data via Slack.
        """
        if greeting_message == DEFAULT_GREETING_MESSAGE:
            greeting_message.format(metric_name=self.metric_name)

        summary_entries = []
        summary_entries.append(greeting_message)

        for _, row in prediction.iterrows():
            date = row[DS_COL].strftime('%Y-%b-%d')
            value = "{:.2f}".format(row[YHAT_COL])
            summary_entries.append(f"• *[{date}]* {value}\n")

        summary_entries.append(farewell_message)

        summary_message = "\n".join(summary_entries)  # type: ignore

        return self.slack_client.files_upload(
            channels=self.channel_id,
            file=str(plot_filename),
            initial_comment=summary_message,
            title=f"{self.metric_name} Forecast",
        )


class SlackReportTask(Step, SlackReport):
    """
    Builds up the task of the report designed for Slack.
    """

    def __init__(
        self,
        channel_id: str,
        metric_name: str,
        setting_path: Optional[str],
        **kwargs: Any,
    ):
        """
        Parameters
        ----------
        channel_id: str
            Slack channel id where the report will be sent.
        metric_name: str
            Performance metric being measured.
        setting_path: str
            Setting path.
        kwargs:
            Extra args to pass.
        """
        Step.__init__(self, **kwargs)  # type: ignore
        SlackReport.__init__(
            self, channel_id, metric_name, setting_path,
        )

    def run(  # type: ignore
        self,
        prediction: pd.DataFrame,
        plot_filename: Union[str, Path],
        greeting_message: Optional[str] = DEFAULT_GREETING_MESSAGE,
        farewell_message: Optional[str] = DEFAULT_FAREWELL_MESSAGE,
    ):
        """
        Send Slack report task.

        Parameters
        ----------
        prediction : pd.DataFrame
            DataDrame of the predictions made.
        plot_filename : str or pathlib.Path
             Path of the forecast data to send.
        greeting_message: str
            Greeting message to send via Slack with the predictions.
        farewell_message: str
            Farewell message to send via Slack with the predictions.

        Returns
        -------
        Slack Report Task
            Sends the specified message with the predictions data via Slack.
        """
        return self.send_report(
            prediction, plot_filename, greeting_message, farewell_message
        )



from unittest.mock import MagicMock, patch

from pandas.testing import assert_frame_equal
import pytest

from soam.models import SkOrbit
from soam.utilities.utils import add_future_dates
from tests.helpers import sample_data_df  # pylint: disable=unused-import


def test_fit(sample_data_df):  # pylint: disable=redefined-outer-name
    with patch("soam.models._orbit.DLTFull") as model_patch:
        output_length = 10
        train_data = sample_data_df
        data = add_future_dates(train_data, output_length)
        X, y = data[data.columns[:-1]], data[data.columns[-1]]
        wrapper = SkOrbit()
        model_patch.assert_not_called()
        wrapper.fit(X, y)
        model_patch.assert_called_once()
        fit_call_df = model_patch.return_value.fit.call_args[0][0]
        assert_frame_equal(data, fit_call_df)


def test_predict_without_fit_fails(
    sample_data_df,
):  # pylint: disable=redefined-outer-name
    with patch("soam.models._orbit.DLTFull") as model_patch:
        output_length = 10
        train_data = sample_data_df
        data = add_future_dates(train_data, output_length)
        X = data[data.columns[:-1]]
        wrapper = SkOrbit()
        model_patch.assert_not_called()
        with pytest.raises(AttributeError):
            wrapper.predict(X)


def test_fit_transform(sample_data_df):  # pylint: disable=redefined-outer-name
    with patch("soam.models._orbit.DLTFull") as model_patch:
        output_length = 10
        train_data = sample_data_df
        data = add_future_dates(train_data, output_length)
        X, y = data[data.columns[:-1]], data[data.columns[-1]]
        wrapper = SkOrbit()
        model_patch.assert_not_called()
        wrapper.fit_transform(X, y)
        model_patch.assert_called_once()
        model_patch.return_value.fit.assert_called_once()
        model_patch.return_value.predict.assert_called_once()


def test_predict():
    X = MagicMock()
    wrapper = SkOrbit()
    model_mock = MagicMock()
    wrapper.model = model_mock
    wrapper.predict(X)
    model_mock.predict.assert_called_once_with(X)
    model_mock.predict.return_value.rename.assert_called_once_with(
        columns={"prediction": "yhat"}
    )



"""
Google Sheets report generation
-------------------------------
GSheets report creator and sender. Its a postprocess that creates a GSheets report with
the model forecasts.
"""
import logging
from pathlib import Path
from typing import Mapping

from muttlib.gsheetsconn import GSheetsClient
import pandas as pd

from soam.core import Step

logger = logging.getLogger(__name__)


class GSheetsReportTask(Step):
    """
    Creates the task to generate a GSheets report.
    """

    def __init__(
        self, config_json_path: str, gsheets_kwargs: Mapping = None, **kwargs
    ):  # pylint: disable=dangerous-default-value
        """
        Parameters
        ----------
        config_json_path: str
            Path to GSheets config json
        gsheets_kwargs: dict
            Extra args to pass to muttlib.gsheetsconn.GSheetsClient
        kwargs:
            Extra args to pass to soam.core.Step
        """
        Step.__init__(self, **kwargs)  # type: ignore
        config_path = Path(config_json_path)
        if not config_path.is_file():
            raise ValueError("Config path does not point to a file.")
        if not gsheets_kwargs:
            gsheets_kwargs = {}
        self.client = GSheetsClient(config_path, **gsheets_kwargs)

    def run(self, df: pd.DataFrame, spreadsheet: str, **insert_kwargs):  # type: ignore[override]
        """
        Save df to a GSheets spreadsheet

        Parameters
        ----------
        df: pd.DataFrame
            A pandas DataFrame containing the data for the step
        spreadsheet: str
            Spreadsheet id or name as in Drive.
        kwargs:
            Extra args to pass to muttlib.gsheetsconn.GSheetsClient.insert_from_frame

        Returns
        -------
        GSheets Report Task
            Google Sheets spreadsheet with the df values.
        """

        if df.empty:
            logger.warning("Exporting empty DataFrame")

        return self.client.insert_from_frame(df, spreadsheet, **insert_kwargs)



"""Forecast plotter tests."""
from copy import deepcopy
import locale

import numpy as np
import pytest

from soam.constants import (
    ANOMALY_PLOT,
    FIG_SIZE,
    MONTHLY_TIME_GRANULARITY,
    PLOT_CONFIG,
    Y_COL,
    YHAT_COL,
)
from soam.plotting import ForecastPlotterTask
from tests.helpers import sample_data_df  # pylint: disable=unused-import


@pytest.fixture
def set_time_locale():
    locale.setlocale(locale.LC_TIME, (None, None))


def perturb_ts(df, col, scale=1):
    """Add noise to ts
    """
    mean = df[col].mean() * scale
    df[col] += np.random.default_rng(42).uniform(
        low=-mean / 2, high=mean / 2, size=len(df)
    )
    return df


def assert_out_paths_equal(fnames, tmp_path):
    expected_out_files = [(tmp_path / fn).as_posix() for fn in fnames]
    out_files = [p.as_posix() for p in tmp_path.iterdir()]
    assert len(expected_out_files) == len(out_files)
    assert all(a == b for a, b in zip(expected_out_files, out_files))


def run_standard_ForecastPlotterTask(tmp_path, time_series, prediction):
    plot_config = deepcopy(PLOT_CONFIG)
    plot_config[ANOMALY_PLOT][MONTHLY_TIME_GRANULARITY][FIG_SIZE] = (8, 3)
    fpt = ForecastPlotterTask(
        path=tmp_path,
        metric_name='test',
        time_granularity=MONTHLY_TIME_GRANULARITY,
        plot_config=plot_config,
    )
    fpt.run(time_series, prediction)
    return fpt


@pytest.mark.mpl_image_compare
def test_ForecastPlotterTask_simple(
    tmp_path, sample_data_df, set_time_locale
):  # pylint: disable=redefined-outer-name,unused-argument
    time_series = sample_data_df.iloc[:30]
    prediction = sample_data_df.iloc[30:]
    prediction = prediction.rename(columns={Y_COL: YHAT_COL})
    fpt = run_standard_ForecastPlotterTask(tmp_path, time_series, prediction)
    assert_out_paths_equal(['0_forecast_2013020100_2015080100_.png'], tmp_path)
    return fpt.fig


@pytest.mark.mpl_image_compare
def test_ForecastPlotterTask_overlapping(
    tmp_path, sample_data_df, set_time_locale
):  # pylint: disable=redefined-outer-name,unused-argument
    time_series = sample_data_df
    prediction = sample_data_df.iloc[30:]
    prediction = prediction.rename(columns={Y_COL: YHAT_COL})
    prediction = perturb_ts(prediction, YHAT_COL, scale=0.1)
    fpt = run_standard_ForecastPlotterTask(tmp_path, time_series, prediction)
    assert_out_paths_equal(['0_forecast_2013020100_2015080100_.png'], tmp_path)
    return fpt.fig
